// Copyright (C) 2006 Pushok Software http://www.pushok.com/soft_svncom.php
//
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html
//
// Subversion is Copyrght of(c) 2000-2004 CollabNet.  All rights reserved.
// Full license available here http://subversion.tigris.org/license-1.html.

#pragma once
#include "resource.h"       // main symbols
#include <vector>
#include "pcom.h"

#define LOCK_THIS_OBJ ObjectLock Lock(this);

namespace NStringVector
{
	// Store our data in a vector of std::strings
	typedef std::vector< CComBSTR >					ContainerType;

	// Now we have all the information we need to fill in the template arguments on the implementation classes
	typedef CComEnumOnSTL< IEnumVARIANT, &__uuidof(IEnumVARIANT), VARIANT,
							PushokAtlEx::GenericCopy<VARIANT, BSTR>, ContainerType > EnumeratorType;

	typedef ICollectionOnSTLImpl< IStringVector, ContainerType, BSTR,
							PushokAtlEx::GenericCopy<BSTR>, EnumeratorType > CollectionType;
};

class ATL_NO_VTABLE CStringVector : 
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CStringVector, &CLSID_StringVector>,
	public IDispatchImpl<NStringVector::CollectionType,&__uuidof(IStringVector)>
{
public:
	CStringVector()
	{
	}

DECLARE_REGISTRY_RESOURCEID(IDR_STRINGVECTOR)

BEGIN_COM_MAP(CStringVector)
	COM_INTERFACE_ENTRY(IStringVector)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY_AGGREGATE(IID_IMarshal, m_pUnkMarshaler.p)
END_COM_MAP()


	DECLARE_PROTECT_FINAL_CONSTRUCT()
	DECLARE_GET_CONTROLLING_UNKNOWN()

	HRESULT FinalConstruct()
	{
		return CoCreateFreeThreadedMarshaler(
			GetControllingUnknown(), &m_pUnkMarshaler.p);
	}
	
	void FinalRelease() 
	{
		m_pUnkMarshaler.Release();
	}

	CComPtr<IUnknown> m_pUnkMarshaler;

public:
	STDMETHOD(Add)(BSTR pVal) {
		LOCK_THIS_OBJ
		m_coll.push_back(CComBSTR(pVal));
		return S_OK;
	}
	STDMETHOD(Remove) (long Index) {
		LOCK_THIS_OBJ
		try {
			m_coll.erase(m_coll.begin() + Index);
		} catch(...) {
			return E_FAIL;
		}
		return S_OK;
	}
	STDMETHOD(Clear) (void) {
		LOCK_THIS_OBJ
		m_coll.clear();
		return S_OK;
	}
};

OBJECT_ENTRY_AUTO(__uuidof(StringVector), CStringVector)

namespace PushokAtlEx
{
	template <>
	class GenericCopy<BSTR,CComBSTR>
	{
	public :
		typedef BSTR	destination_type;
		typedef CComBSTR	source_type;

		static void init(destination_type* p)
		{
			*p = NULL;
		}
		static void destroy(destination_type* p)
		{
			SysFreeString(*p);
			*p = NULL;
		}
		static HRESULT copy(destination_type* pTo, const CComBSTR* pFrom)
		{
			return const_cast<CComBSTR*>(pFrom)->CopyTo(pTo);
		}
	}; // class GenericCopy<BSTR>
	template <>
	class GenericCopy<VARIANT, CComBSTR>
	{
	public :
		typedef VARIANT	destination_type;
		typedef CComBSTR	source_type;

		static void init(destination_type* p)
		{
			GenericCopy<destination_type>::init(p);
		}
		static void destroy(destination_type* p)
		{
			GenericCopy<destination_type>::destroy(p);
		}
		static HRESULT copy(destination_type* pTo, const source_type* pFrom)
		{
			return CComVariant(*pFrom).Detach(pTo);
		}

	}; // class GenericCopy<VARIANT, BSTR>
}

namespace NStringMap
{
	// Store our data in a vector of std::strings
	typedef std::map< CComBSTR, CComBSTR >	ContainerType;

	// Now we have all the information we need to fill in the template arguments on the implementation classes
	typedef CComEnumOnSTL< IEnumVARIANT, &__uuidof(IEnumVARIANT), VARIANT,
		PushokAtlEx::MapCopy<ContainerType,VARIANT>, ContainerType > EnumeratorType;

	typedef PushokAtlEx::ICollectionOnSTLCopyImpl< IStringMap, ContainerType, BSTR,
							PushokAtlEx::MapCopy<ContainerType,BSTR>, EnumeratorType > CollectionType;

};

class ATL_NO_VTABLE CStringMap : 
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CStringMap, &CLSID_StringMap>,
	public IDispatchImpl<NStringMap::CollectionType,&__uuidof(IStringMap)>
{
public:
	CStringMap()
	{
	}

DECLARE_REGISTRY_RESOURCEID(IDR_STRINGMAP)

BEGIN_COM_MAP(CStringMap)
	COM_INTERFACE_ENTRY(IStringMap)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY_AGGREGATE(IID_IMarshal, m_pUnkMarshaler.p)
END_COM_MAP()


	DECLARE_PROTECT_FINAL_CONSTRUCT()
	DECLARE_GET_CONTROLLING_UNKNOWN()

	HRESULT FinalConstruct()
	{
		return CoCreateFreeThreadedMarshaler(
			GetControllingUnknown(), &m_pUnkMarshaler.p);
	}
	
	void FinalRelease() 
	{
		m_pUnkMarshaler.Release();
	}

	CComPtr<IUnknown> m_pUnkMarshaler;

public:
	STDMETHOD(get_Item)(BSTR Index, BSTR* pVal)	{
		LOCK_THIS_OBJ
		NStringMap::ContainerType::iterator pIt;
		pIt = m_coll.find(CComBSTR(Index));
		return pIt!=m_coll.end()?(*pIt).second.CopyTo(pVal):E_INVALIDARG;
	}
	STDMETHOD(Add)(BSTR pKey, BSTR pVal) {
		LOCK_THIS_OBJ
		m_coll[pKey] = pVal;
		return S_OK;
	}
	STDMETHOD(Remove) (BSTR pKey) {
		LOCK_THIS_OBJ
//		m_coll. RemovAt(pKey);
		return S_OK;
	}
};

OBJECT_ENTRY_AUTO(__uuidof(StringMap), CStringMap)
